# -*- coding: utf-8 -*-
import sys

import six

from collections import OrderedDict
from functools import wraps

if six.PY3:  # pragma: no branch
    from inspect import signature
else:
    from funcsigs import signature

from six import string_types


class GuardError(Exception):
    pass


def _params(f, argumets, keyword_argumets, params):
    sig = signature(f)
    bind = sig.bind(*argumets, **keyword_argumets)
    if sys.version_info >= (3, 5):
        bind.apply_defaults()

    return OrderedDict(
        filter(
            lambda item: item[0] in params,
            bind.arguments.items()
        )
    )


def guard(params, guardian, error_class=GuardError, message=''):
    '''
    A guard function - check parameters
    with guardian function on decorated function

    :param tuple or string params: guarded function parameter/s
    :param function guardian: verifying the conditions for the selected parameter
    :param Exception error_class: raised class when guardian return false
    :param string message: error message
    '''
    params = [params] if isinstance(params, string_types) else params

    def guard_decorate(f):
        @wraps(f)
        def _guard_decorate(*args, **kwargs):
            if guardian(**_params(f, args, kwargs, params)):
                return f(*args, **kwargs)
            else:
                raise error_class(message)
        return _guard_decorate
    return guard_decorate



__author__ = 'Jindrich K. Smitka'
__email__ = 'smitka.j@gmail.com'
__version__ = '0.4.0'

from .railroad import (  # noqa
    prepare,
    catch,
    get_or_reraise,
    compose,
)
from .rescue import (  # noqa
    reraise,
    nop,
    rescue,
)
from .actions import actions, lift, result  # noqa

from .guard import guard, GuardError  # noqa



# -*- coding: utf-8 -*-

import pytest

from mock import Mock, call

from railroad import guard, GuardError


def fn(a, b, c, d='foo'):
    return (a, b, c, d)


def test_guard_call_guarded_function():
    def fn(a):
        return a

    guarded = guard('a', lambda a: True)(fn)
    result = guarded('foo')

    assert result == 'foo'


def test_guard_pass_only_selected_paramters_to_guardian():
    guardian = Mock()

    guarded = guard(['b', 'a'], guardian)(fn)
    guarded(1, 2, 3, 4)

    assert guardian.called
    assert guardian.call_args == call(a=1, b=2)


def test_guard_raise_exception_when_guardian_return_false():
    def fn(a):
        return a

    guarded = guard('a', lambda a: False)(fn)

    with pytest.raises(GuardError):
        guarded('foo')


def test_guard_raise_custom_exception():
    def fn(a):
        return a

    class MyError(Exception):
        pass

    guarded = guard('a', lambda a: False, MyError)(fn)

    with pytest.raises(MyError):
        guarded('foo')


def test_guard_called_with_pramas_as_string():
    def fn(a, ab):
        return ab

    guarded = guard('ab', lambda ab: True)(fn)
    result = guarded(1, 'foo')

    assert result == 'foo'



# -*- coding: utf-8 -*-

from functools import wraps, partial
from six.moves import reduce
from toolz import concatv, cons, remove


def actions(acts, done):
    '''
    Prepare actions pipeline.

    :param tuple acts: called functions
    :param function done: get result from actions
    :returns function: function that starts executio
    '''

    def _intermediate(acc, action):
        result = action(acc['state'])
        values = concatv(acc['values'], [result['answer']])

        return {'values': values, 'state': result['state']}

    def _actions(seed):
        init = {'values': [], 'state': seed}

        result = reduce(_intermediate, acts, init)

        keep = remove(lambda x: x is None, result['values'])

        return done(keep, result['state'])

    return _actions


def lift(fn=None, state_fn=None):
    """
    The lift decorator function will be used to abstract away the management
    of the state object used as the intermediate representation of actions.

    :param function answer: a function to provide
                            the result of some action given a value
    :param function state: a function to provide what the new state looks like
    :returns function: a function suitable for use in actions
    """
    if fn is None:
        return partial(lift, state_fn=state_fn)

    @wraps(fn)
    def _lift(*args, **kwargs):
        def _run(state):
            ans = fn(*cons(state, args), **kwargs)
            s = state_fn(state) if state_fn is not None else ans

            return {'answer': ans, 'state': s}
        return _run
    return _lift


def result(values, state):
    '''
    Return computed value from actions
    '''
    return state



# -*- coding: utf-8 -*-

from random import randint

from mock import Mock, call
from toolz import cons

from railroad import actions, lift, result


def test_result_return_state():
    r = result([1, 2, 3], 'state')

    assert r == 'state'


def test_run_lifted_function_return_intermediate_state_crate():
    f = Mock()
    f.__name__ = 'f'
    lifted = lift(f)
    prepered = lifted()

    result = prepered('foo')

    assert result['answer'] == f.return_value
    assert result['state'] == f.return_value


def test_lift_call_state_fn():
    f = Mock()
    f.__name__ = 'f'
    state_fn = Mock()
    lifted = lift(state_fn=state_fn)(f)
    prepered = lifted()

    prepered('foo')

    assert state_fn.called
    assert state_fn.call_args == call('foo')


def test_actions_call_done_with_state_and_values():

    def f(a, b):
        return '%s%d' % (a, b)

    fl = lift(f)
    instructions = [fl(i) for i in range(randint(1, 1000))]

    run = actions(instructions, result)
    r = run('seed')

    expect = ''.join(cons('seed', [str(i) for i in range(len(instructions))]))

    assert r == expect



# -*- coding: utf-8 -*-

import pytest

from mock import Mock, call

from railroad import rescue


def test_rescue_call_original_function():
    on_success = Mock()
    f = Mock()

    rescue(f, on_success)(1, 2, 'foo')  # call f function

    assert f.called
    assert f.call_args == call(1, 2, 'foo')


def test_rescue_call_on_success_with_function_return_value_as_parameter():
    on_success = Mock()
    f = Mock()

    rescue(f, on_success)()  # call f function

    assert on_success.called
    assert on_success.call_args == call(f.return_value)


def test_rescue_return_on_success_result():
    on_success = Mock()
    f = Mock()

    result = rescue(f, on_success)()  # call f function

    assert result == on_success.return_value


def test_rescue_in_default_reraise_exception():
    on_success = Mock()
    f = Mock()

    class MyException(Exception):
        pass

    f.side_effect = MyException

    with pytest.raises(MyException):
        rescue(f, on_success)()  # call f function


def test_rescue_call_on_error_when_exception():
    on_success = Mock()
    f = Mock()
    f.side_effect = Exception

    def on_error(e):
        assert e.__class__ == Exception

    rescue(f, on_success, on_error)()  # call f function


def test_rescue_return_on_error_value():
    on_success = Mock()
    on_error = Mock()
    f = Mock()
    f.side_effect = Exception

    result = rescue(f, on_success, on_error)()  # call f function

    assert result == on_error.return_value


def test_rescue_call_on_complete():
    on_complete = Mock()

    rescue(Mock(), Mock(), Mock(), on_complete)()  # call f function

    assert on_complete.called



"""
Tests for `railroad` module.
"""
import pytest

from mock import Mock, call

from railroad import (
    prepare,
    catch,
    get_or_reraise,
    compose,
)


def test_get_or_reraise_return_data_if_has_no_error():
    data = ('test', None)

    result = get_or_reraise(data)

    assert result == 'test'


def test_get_or_reraise_reraise_given_errror():
    data = (None, Exception())
    with pytest.raises(Exception):
        get_or_reraise(data)


def test_call_compose_with_right_order():
    f = compose(lambda x: x + 'a', lambda x: x + 'b')
    assert f('') == 'ab'


def test_catch_given_error_and_call_function():
    ex = Exception()
    data = (None, ex)
    f = Mock()

    c = catch(Exception, f)

    result = c(data)

    assert f.called
    assert f.call_args == call(ex)

    assert result[0] == f.return_value
    assert result[1] is None


def test_catch_return_data_when_no_error():
    data = ('test', None)

    f = Mock()

    c = catch(Exception, f)

    result = c(data)

    assert not f.called

    assert result == data


def test_prepare_catch_exception_and_return_it_in_tuple():
    f = Mock(side_effect=Exception)
    fn = prepare(f)

    result = fn()

    assert result[0] is None
    assert isinstance(result[1], Exception)


def test_prepare_return_result_from_given_function_in_tuple():
    f = Mock()
    fn = prepare(f)

    result = fn()

    assert result[0] == f.return_value
    assert result[1] is None



#!/usr/bin/env python

import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

module_path = os.path.join(os.path.dirname(__file__), 'railroad', '__init__.py')
version_line = [line for line in open(module_path)
                if line.startswith('__version__')][0]
__version__ = eval(version_line.split('__version__ = ')[-1])

readme = open('README.rst').read()
doclink = """
Documentation
-------------

"""
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='railroad',
    version=__version__,
    description='A functional library for data processing.',
    long_description=readme + '\n\n' + doclink + '\n\n' + history,
    author='Jindrich K. Smitka',
    author_email='smitka.j@gmail.com',
    url='https://github.com/s-m-i-t-a/railroad',
    packages=[
        'railroad',
    ],
    package_dir={'railroad': 'railroad'},
    include_package_data=True,
    install_requires=[
        'six>=1.7.3',
        'toolz>=0.7.4',
        ' funcsigs>=1.0.2',
    ],
    license='MIT',
    zip_safe=False,
    keywords='railroad',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)



# -*- coding: utf-8 -*-

from toolz import compose as c


def prepare(fn):
    def _prepare(*args, **kwargs):
        try:
            return (fn(*args, **kwargs), None)
        except Exception as e:
            return (None, e)

    return _prepare


def catch(error, fn):
    def _catch(container):
        data, excp = container

        if isinstance(excp, error):
            return (fn(excp), None)
        else:
            return container

    return _catch


def get_or_reraise(container):
    data, excp = container

    if excp is None:
        return data
    else:
        raise excp


def compose(*args):
    return c(*reversed(args))



# -*- coding: utf-8 -*-


def reraise(e):
    raise e


def nop(*args, **kwargs):
    pass


def rescue(f, on_success, on_error=reraise, on_complete=nop):
    '''
    Functional try-except-finally

    :param function f: guarded function
    :param function on_succes: called when f is executed without error
    :param function on_error: called with `error` parameter when f failed
    :param function on_complete: called as finally block
    :returns function: call signature is equal f signature
    '''
    def _rescue(*args, **kwargs):
        try:
            return on_success(f(*args, **kwargs))
        except Exception as e:
            return on_error(e)
        finally:
            on_complete()

    return _rescue
